@@ -3,7 +3,6 @@
 from __future__ import absolute_import, print_function
 
 import argparse
-import collections
 import copy
 import functools
 import io
@@ -22,23 +21,28 @@ import pkg_resources  # part of setuptools
 from ruamel import yaml
 from ruamel.yaml.comments import CommentedMap, CommentedSeq
 from schema_salad import validate
-from schema_salad.ref_resolver import Loader, file_uri, uri_file_path
+from schema_salad.ref_resolver import Fetcher, Loader, file_uri, uri_file_path
 from schema_salad.sourceline import strip_dup_lineno, cmap
 from six import string_types, iteritems, PY3
 from typing_extensions import Text
 # move to a regular typing import when Python 3.3-3.6 is no longer supported
 
+if PY3:
+  from collections.abc import Iterable, Sequence, MutableSequence
+else:  # Needed for Py3.8
+  from collections import Iterable, Sequence, MutableSequence
+
 from . import command_line_tool, workflow
 from .argparser import arg_parser, generate_parser, get_default_args
 from .builder import HasReqsHints  # pylint: disable=unused-import
 from .context import LoadingContext, RuntimeContext, getdefault
 from .cwlrdf import printdot, printrdf
 from .errors import UnsupportedRequirement, WorkflowException
-from .executors import MultithreadedJobExecutor, SingleJobExecutor
+from .executors import MultithreadedJobExecutor, SingleJobExecutor, JobExecutor
 from .load_tool import (FetcherConstructorType,  # pylint: disable=unused-import
                         fetch_document, jobloaderctx, load_overrides,
                         make_tool, resolve_overrides, resolve_tool_uri,
-                        validate_document)
+                        resolve_and_validate_document, default_loader)
 from .loghandler import _logger, defaultStreamHandler
 from .mutation import MutationManager
 from .pack import pack
@@ -59,6 +63,8 @@ from .utils import (DEFAULT_TMP_PREFIX, json_dumps, onWindows,
                     windows_default_container_id)
 from .subgraph import get_subgraph
 
+import coloredlogs
+
 def _terminate_processes():
     # type: () -> None
     """Kill all spawned processes.
@@ -93,7 +99,7 @@ def _signal_handler(signum, _):
 def generate_example_input(inptype,     # type: Any
                            default      # type: Optional[Any]
                           ):  # type: (...) -> Tuple[Any, Text]
-    """Converts a single input schema into an example."""
+    """Convert a single input schema into an example."""
     example = None
     comment = u""
     defaults = {u'null': 'null',
@@ -109,7 +115,7 @@ def generate_example_input(inptype,     # type: Any
                 u'Directory': yaml.comments.CommentedMap([
                     ('class', 'Directory'), ('path', 'a/directory/path')])
                }  # type: Dict[Text, Any]
-    if isinstance(inptype, collections.MutableSequence):
+    if isinstance(inptype, MutableSequence):
         optional = False
         if 'null' in inptype:
             inptype.remove('null')
@@ -129,7 +135,7 @@ def generate_example_input(inptype,     # type: Any
                 example.yaml_add_eol_comment(e_comment, index)
             if optional:
                 comment = u"optional"
-    elif isinstance(inptype, collections.Mapping) and 'type' in inptype:
+    elif isinstance(inptype, Mapping) and 'type' in inptype:
         if inptype['type'] == 'array':
             if len(inptype['items']) == 1 and 'type' in inptype['items'][0] \
                     and inptype['items'][0]['type'] == 'enum':
@@ -191,16 +197,16 @@ def realize_input_schema(input_types,  # type: MutableSequence[Dict[Text, Any]]
                 input_type_name = entry
             if input_type_name in schema_defs:
                 entry = input_types[index] = schema_defs[input_type_name]
-        if isinstance(entry, collections.Mapping):
+        if isinstance(entry, Mapping):
             if isinstance(entry['type'], string_types) and '#' in entry['type']:
                 _, input_type_name = entry['type'].split('#')
                 if input_type_name in schema_defs:
                     input_types[index]['type'] = realize_input_schema(
                         schema_defs[input_type_name], schema_defs)
-            if isinstance(entry['type'], collections.MutableSequence):
+            if isinstance(entry['type'], MutableSequence):
                 input_types[index]['type'] = realize_input_schema(
                     entry['type'], schema_defs)
-            if isinstance(entry['type'], collections.Mapping):
+            if isinstance(entry['type'], Mapping):
                 input_types[index]['type'] = realize_input_schema(
                     [input_types[index]['type']], schema_defs)
             if entry['type'] == 'array':
@@ -225,7 +231,7 @@ def generate_input_template(tool):
 
 def load_job_order(args,                 # type: argparse.Namespace
                    stdin,                # type: IO[Any]
-                   fetcher_constructor,  # Fetcher
+                   fetcher_constructor,  # type: Optional[Fetcher]
                    overrides_list,       # type: List[Dict[Text, Any]]
                    tool_file_uri         # type: Text
                   ):  # type: (...) -> Tuple[Optional[MutableMapping[Text, Any]], Text, Loader]
@@ -279,7 +285,7 @@ def init_job_order(job_order_object,        # type: Optional[MutableMapping[Text
                    relative_deps=False,     # type: bool
                    make_fs_access=StdFsAccess,  # type: Callable[[Text], StdFsAccess]
                    input_basedir="",        # type: Text
-                   secret_store=None        # type: SecretStore
+                   secret_store=None        # type: Optional[SecretStore]
                   ):  # type: (...) -> MutableMapping[Text, Any]
     secrets_req, _ = process.get_requirement("http://commonwl.org/cwltool#Secrets")
     if job_order_object is None:
@@ -300,11 +306,11 @@ def init_job_order(job_order_object,        # type: Optional[MutableMapping[Text
                 record[key[len(record_name) + 1:]] = value
                 del cmd_line[key]
             cmd_line[str(record_name)] = record
-
         if 'job_order' in cmd_line and cmd_line["job_order"]:
             try:
                 job_order_object = cast(
-                    MutableMapping, loader.resolve_ref(cmd_line["job_order"])[0])
+                    MutableMapping[Text, Any],
+                    loader.resolve_ref(cmd_line["job_order"])[0])
             except Exception as err:
                 _logger.error(Text(err), exc_info=args.debug)
                 exit(1)
@@ -342,7 +348,7 @@ def init_job_order(job_order_object,        # type: Optional[MutableMapping[Text
             job_order_object = {}
 
     if print_input_deps:
-        basedir = None
+        basedir = None  # type: Optional[Text]
         uri = job_order_object["id"]
         if uri == args.workflow:
             basedir = os.path.dirname(uri)
@@ -351,16 +357,17 @@ def init_job_order(job_order_object,        # type: Optional[MutableMapping[Text
                   basedir=basedir, nestdirs=False)
         exit(0)
 
-    def path_to_loc(p):
+    def path_to_loc(p):  # type: (Dict[Text, Any]) -> None
         if "location" not in p and "path" in p:
             p["location"] = p["path"]
             del p["path"]
 
     ns = {}  # type: Dict[Text, Union[Dict[Any, Any], Text, Iterable[Text]]]
+    ns.update(job_order_object.get("$namespaces", {}))
     ns.update(process.metadata.get("$namespaces", {}))
     ld = Loader(ns)
 
-    def expand_formats(p):
+    def expand_formats(p):  # type: (Dict[Text, Any]) -> None
         if "format" in p:
             p["format"] = ld.expand_url(p["format"], "")
 
@@ -382,7 +389,7 @@ def init_job_order(job_order_object,        # type: Optional[MutableMapping[Text
 
 
 
-def make_relative(base, obj):
+def make_relative(base, obj):  # type: (Text, Dict[Text, Any]) -> None
     """Relativize the location URI of a File or Directory object."""
     uri = obj.get("location", obj.get("path"))
     if ":" in uri.split("/")[0] and not uri.startswith("file://"):
@@ -397,7 +404,7 @@ def printdeps(obj,              # type: Mapping[Text, Any]
               stdout,           # type: Union[TextIO, StreamWriter]
               relative_deps,    # type: bool
               uri,              # type: Text
-              basedir=None,     # type: Text
+              basedir=None,     # type: Optional[Text]
               nestdirs=True     # type: bool
              ):  # type: (...) -> None
     """Print a JSON representation of the dependencies of the CWL document."""
@@ -414,7 +421,7 @@ def printdeps(obj,              # type: Mapping[Text, Any]
 def prov_deps(obj,              # type: Mapping[Text, Any]
               document_loader,  # type: Loader
               uri,              # type: Text
-              basedir=None      # type: Text
+              basedir=None      # type: Optional[Text]
              ):  # type: (...) -> MutableMapping[Text, Any]
     deps = find_deps(obj, document_loader, uri, basedir=basedir)
 
@@ -434,13 +441,13 @@ def prov_deps(obj,              # type: Mapping[Text, Any]
 def find_deps(obj,              # type: Mapping[Text, Any]
               document_loader,  # type: Loader
               uri,              # type: Text
-              basedir=None,     # type: Text
+              basedir=None,     # type: Optional[Text]
               nestdirs=True     # type: bool
              ):  # type: (...) -> Dict[Text, Any]
     """Find the dependencies of the CWL document."""
     deps = {"class": "File", "location": uri, "format": CWL_IANA}  # type: Dict[Text, Any]
 
-    def loadref(base, uri):
+    def loadref(base, uri):  # type: (Text, Text) -> Any
         return document_loader.fetch(document_loader.fetcher.urljoin(base, uri))
 
     sfs = scandeps(
@@ -452,10 +459,10 @@ def find_deps(obj,              # type: Mapping[Text, Any]
     return deps
 
 def print_pack(document_loader,  # type: Loader
-               processobj,       # type: Union[Dict[Text, Any], List[Dict[Text, Any]]]
+               processobj,       # type: CommentedMap
                uri,              # type: Text
                metadata          # type: Dict[Text, Any]
-              ):  # type (...) -> Text
+              ):  # type: (...) -> Text
     """Return a CWL serialization of the CWL document in JSON."""
     packed = pack(document_loader, processobj, uri, metadata)
     if len(packed["$graph"]) > 1:
@@ -472,21 +479,21 @@ def supported_cwl_versions(enable_dev):  # type: (bool) -> List[Text]
     versions.sort()
     return versions
 
-def main(argsl=None,                   # type: List[str]
-         args=None,                    # type: argparse.Namespace
-         job_order_object=None,        # type: MutableMapping[Text, Any]
+def main(argsl=None,                   # type: Optional[List[str]]
+         args=None,                    # type: Optional[argparse.Namespace]
+         job_order_object=None,        # type: Optional[MutableMapping[Text, Any]]
          stdin=sys.stdin,              # type: IO[Any]
-         stdout=None,                  # type: Union[TextIO, StreamWriter]
+         stdout=None,                  # type: Optional[Union[TextIO, StreamWriter]]
          stderr=sys.stderr,            # type: IO[Any]
          versionfunc=versionstring,    # type: Callable[[], Text]
-         logger_handler=None,          #
-         custom_schema_callback=None,  # type: Callable[[], None]
-         executor=None,                # type: Callable[..., Tuple[Dict[Text, Any], Text]]
-         loadingContext=None,          # type: LoadingContext
-         runtimeContext=None           # type: RuntimeContext
+         logger_handler=None,          # type: Optional[logging.Handler]
+         custom_schema_callback=None,  # type: Optional[Callable[[], None]]
+         executor=None,                # type: Optional[JobExecutor]
+         loadingContext=None,          # type: Optional[LoadingContext]
+         runtimeContext=None           # type: Optional[RuntimeContext]
         ):  # type: (...) -> int
     if not stdout:  # force UTF-8 even if the console is configured differently
-        if (hasattr(sys.stdout, "encoding")  # type: ignore
+        if (hasattr(sys.stdout, "encoding")
                 and sys.stdout.encoding != 'UTF-8'):  # type: ignore
             if PY3 and hasattr(sys.stdout, "detach"):
                 stdout = io.TextIOWrapper(sys.stdout.buffer, encoding='utf-8')
@@ -496,12 +503,12 @@ def main(argsl=None,                   # type: List[str]
             stdout = cast(TextIO, sys.stdout)  # type: ignore
 
     _logger.removeHandler(defaultStreamHandler)
-    if logger_handler is not None:
-        stderr_handler = logger_handler
+    stderr_handler = logger_handler
+    if stderr_handler is not None:
+        _logger.addHandler(stderr_handler)
     else:
-        stderr_handler = logging.StreamHandler(stderr)
-    _logger.addHandler(stderr_handler)
-    # pre-declared for finally block
+        coloredlogs.install(logger=_logger, stream=stderr)
+        stderr_handler = _logger.handlers[-1]
     workflowobj = None
     prov_log_handler = None  # type: Optional[logging.StreamHandler]
     try:
@@ -543,12 +550,15 @@ def main(argsl=None,                   # type: List[str]
         if runtimeContext.debug:
             # Increase to debug for both stderr and provenance log file
             _logger.setLevel(logging.DEBUG)
+            stderr_handler.setLevel(logging.DEBUG)
             rdflib_logger.setLevel(logging.DEBUG)
-        formatter = None  # type: Optional[logging.Formatter]
+        fmtclass = coloredlogs.ColoredFormatter if args.enable_color else logging.Formatter
+        formatter = fmtclass("%(levelname)s %(message)s")
         if args.timestamps:
-            formatter = logging.Formatter("[%(asctime)s] %(message)s",
-                                          "%Y-%m-%d %H:%M:%S")
-            stderr_handler.setFormatter(formatter)
+            formatter = fmtclass(
+                "[%(asctime)s] %(levelname)s %(message)s",
+                "%Y-%m-%d %H:%M:%S")
+        stderr_handler.setFormatter(formatter)
         ##
 
         if args.version:
@@ -564,7 +574,6 @@ def main(argsl=None,                   # type: List[str]
             if os.path.isfile("CWLFile"):
                 setattr(args, "workflow", "CWLFile")
             else:
-                _logger.error("")
                 _logger.error("CWL document required, no input file was provided")
                 arg_parser().print_help()
                 return 1
@@ -593,16 +602,17 @@ def main(argsl=None,                   # type: List[str]
                 full_name=args.cwl_full_name)
             runtimeContext.research_obj = ro
             log_file_io = ro.open_log_file_for_activity(ro.engine_uuid)
-            prov_log_handler = logging.StreamHandler(log_file_io)
+            prov_log_handler = logging.StreamHandler(cast(IO[str], log_file_io))
 
             class ProvLogFormatter(logging.Formatter):
                 """Enforce ISO8601 with both T and Z."""
+
                 def __init__(self):  # type: () -> None
                     super(ProvLogFormatter, self).__init__(
                         "[%(asctime)sZ] %(message)s")
 
                 def formatTime(self, record, datefmt=None):
-                    # type: (logging.LogRecord, str) -> str
+                    # type: (logging.LogRecord, Optional[str]) -> str
                     record_time = time.gmtime(record.created)
                     formatted_time = time.strftime("%Y-%m-%dT%H:%M:%S", record_time)
                     with_msecs = "%s,%03d" % (formatted_time, record.msecs)
@@ -619,12 +629,15 @@ def main(argsl=None,                   # type: List[str]
             loadingContext = LoadingContext(vars(args))
         else:
             loadingContext = loadingContext.copy()
+        loadingContext.loader = default_loader(loadingContext.fetcher_constructor)
         loadingContext.research_obj = runtimeContext.research_obj
         loadingContext.disable_js_validation = \
             args.disable_js_validation or (not args.do_validate)
         loadingContext.construct_tool_object = getdefault(
             loadingContext.construct_tool_object, workflow.default_make_tool)
         loadingContext.resolver = getdefault(loadingContext.resolver, tool_resolver)
+        if loadingContext.do_update is None:
+            loadingContext.do_update = not (args.pack or args.print_subgraph)
 
         uri, tool_file_uri = resolve_tool_uri(
             args.workflow, resolver=loadingContext.resolver,
@@ -641,43 +654,40 @@ def main(argsl=None,                   # type: List[str]
                 loadingContext.overrides_list.extend(load_overrides(
                     file_uri(os.path.abspath(args.overrides)), tool_file_uri))
 
-            document_loader, workflowobj, uri = fetch_document(
-                uri, resolver=loadingContext.resolver,
-                fetcher_constructor=loadingContext.fetcher_constructor)
+            loadingContext, workflowobj, uri = fetch_document(
+                uri, loadingContext)
 
-            if args.print_deps:
-                printdeps(workflowobj, document_loader, stdout,
-                           args.relative_deps, uri)
+            if args.print_deps and loadingContext.loader:
+                printdeps(workflowobj, loadingContext.loader, stdout,
+                          args.relative_deps, uri)
                 return 0
 
-            document_loader, avsc_names, processobj, metadata, uri \
-                = validate_document(document_loader, workflowobj, uri,
-                                    loadingContext.overrides_list, {},
-                                    enable_dev=loadingContext.enable_dev,
-                                    strict=loadingContext.strict,
+            loadingContext, uri \
+                = resolve_and_validate_document(loadingContext, workflowobj, uri,
                                     preprocess_only=(args.print_pre or args.pack),
-                                    fetcher_constructor=loadingContext.fetcher_constructor,
-                                    skip_schemas=args.skip_schemas,
-                                    do_validate=loadingContext.do_validate)
+                                    skip_schemas=args.skip_schemas)
+
+            if loadingContext.loader is None:
+                raise Exception("Impossible code path.")
+            processobj, metadata = loadingContext.loader.resolve_ref(uri)
+            processobj = cast(CommentedMap, processobj)
             if args.pack:
-                stdout.write(print_pack(document_loader, processobj, uri, metadata))
+                stdout.write(print_pack(loadingContext.loader, processobj, uri, metadata))
                 return 0
 
             if args.provenance and runtimeContext.research_obj:
                 # Can't really be combined with args.pack at same time
                 runtimeContext.research_obj.packed_workflow(
-                    print_pack(document_loader, processobj, uri, metadata))
+                    print_pack(loadingContext.loader, processobj, uri, metadata))
 
             if args.print_pre:
                 stdout.write(json_dumps(processobj, indent=4, sort_keys=True, separators=(',', ': ')))
                 return 0
 
-            loadingContext.overrides_list.extend(metadata.get("cwltool:overrides", []))
-
-            tool = make_tool(document_loader, avsc_names,
-                             metadata, uri, loadingContext)
+            tool = make_tool(uri, loadingContext)
             if args.make_template:
                 def my_represent_none(self, data):  # pylint: disable=unused-argument
+                    # type: (Any, Any) -> Any
                     """Force clean representation of 'null'."""
                     return self.represent_scalar(u'tag:yaml.org,2002:null', u'null')
                 yaml.RoundTripRepresenter.add_representer(type(None), my_represent_none)
@@ -691,11 +701,11 @@ def main(argsl=None,                   # type: List[str]
                 return 0
 
             if args.print_rdf:
-                stdout.write(printrdf(tool, document_loader.ctx, args.rdf_serializer))
+                stdout.write(printrdf(tool, loadingContext.loader.ctx, args.rdf_serializer))
                 return 0
 
             if args.print_dot:
-                printdot(tool, document_loader.ctx, stdout)
+                printdot(tool, loadingContext.loader.ctx, stdout)
                 return 0
 
             if args.print_targets:
@@ -711,17 +721,18 @@ def main(argsl=None,                   # type: List[str]
                     if url.fragment:
                         extracted = get_subgraph([tool.tool["id"] + "/" + r for r in args.target], tool)
                     else:
-                        extracted = get_subgraph([document_loader.fetcher.urljoin(tool.tool["id"], "#" + r)
+                        extracted = get_subgraph([loadingContext.loader.fetcher.urljoin(tool.tool["id"], "#" + r)
                                                  for r in args.target],
                                                  tool)
                 else:
                     _logger.error("Can only use --target on Workflows")
                     return 1
-                del document_loader.idx[extracted["id"]]
-                tool = make_tool(document_loader, avsc_names,
-                                 metadata,
-                                 cast(CommentedMap, cmap(extracted)),
-                                 loadingContext)
+                if isinstance(loadingContext.loader.idx, CommentedMap):
+                    loadingContext.loader.idx[extracted["id"]] = extracted
+                    tool = make_tool(extracted["id"],
+                                     loadingContext)
+                else:
+                    raise Exception("Missing loadingContext.loader.idx!")
 
             if args.print_subgraph:
                 if "name" in tool.tool:
@@ -730,18 +741,18 @@ def main(argsl=None,                   # type: List[str]
                 return 0
 
         except (validate.ValidationException) as exc:
-            _logger.error(u"Tool definition failed validation:\n%s", exc,
+            _logger.error(u"Tool definition failed validation:\n%s", Text(exc),
                           exc_info=args.debug)
             return 1
         except (RuntimeError, WorkflowException) as exc:
-            _logger.error(u"Tool definition failed initialization:\n%s", exc,
+            _logger.error(u"Tool definition failed initialization:\n%s", Text(exc),
                           exc_info=args.debug)
             return 1
         except Exception as exc:
             _logger.error(
                 u"I'm sorry, I couldn't load this CWL file%s.\nThe error was: %s",
                 try_again_msg,
-                exc if not args.debug else "",
+                Text(exc) if not args.debug else "",
                 exc_info=args.debug)
             return 1
 
@@ -754,6 +765,8 @@ def main(argsl=None,                   # type: List[str]
             default_mac_path = "/private/tmp/docker_tmp"
             if runtimeContext.tmp_outdir_prefix == DEFAULT_TMP_PREFIX:
                 runtimeContext.tmp_outdir_prefix = default_mac_path
+            if runtimeContext.tmpdir_prefix == DEFAULT_TMP_PREFIX:
+                runtimeContext.tmpdir_prefix = default_mac_path
 
         for dirprefix in ("tmpdir_prefix", "tmp_outdir_prefix", "cachedir"):
             if getattr(runtimeContext, dirprefix) and getattr(runtimeContext, dirprefix) != DEFAULT_TMP_PREFIX:
@@ -765,7 +778,7 @@ def main(argsl=None,                   # type: List[str]
                     try:
                         os.makedirs(os.path.dirname(getattr(runtimeContext, dirprefix)))
                     except Exception as e:
-                        _logger.error("Failed to create directory: %s", e)
+                        _logger.error("Failed to create directory: %s", Text(e))
                         return 1
 
         if args.cachedir:
@@ -788,11 +801,13 @@ def main(argsl=None,                   # type: List[str]
 
         if not executor:
             if args.parallel:
-                executor = MultithreadedJobExecutor()
-                runtimeContext.select_resources = executor.select_resources
+                temp_executor = MultithreadedJobExecutor()
+                runtimeContext.select_resources = temp_executor.select_resources
+                real_executor = temp_executor  # type: JobExecutor
             else:
-                executor = SingleJobExecutor()
-        assert executor is not None
+                real_executor = SingleJobExecutor()
+        else:
+            real_executor = executor
 
         try:
             runtimeContext.basedir = input_basedir
@@ -810,17 +825,16 @@ def main(argsl=None,                   # type: List[str]
                     default_container=runtimeContext.default_container,
                     use_biocontainers=args.beta_use_biocontainers)
 
-            (out, status) = executor(tool,
-                                     initialized_job_order_object,
-                                     runtimeContext,
-                                     logger=_logger)
+            (out, status) = real_executor(
+                tool, initialized_job_order_object, runtimeContext,
+                logger=_logger)
 
             if out is not None:
                 if runtimeContext.research_obj is not None:
                     runtimeContext.research_obj.create_job(
                         out, None, True)
 
-                def loc_to_path(obj):
+                def loc_to_path(obj):  # type: (Dict[Text, Any]) -> None
                     for field in ("path", "nameext", "nameroot", "dirname"):
                         if field in obj:
                             del obj[field]
@@ -835,11 +849,10 @@ def main(argsl=None,                   # type: List[str]
                 if isinstance(out, string_types):
                     stdout.write(out)
                 else:
-                    stdout.write(json_dumps(out, indent=4,  # type: ignore
-                                            ensure_ascii=False))
+                    stdout.write(json_dumps(out, indent=4, ensure_ascii=False))
                 stdout.write("\n")
                 if hasattr(stdout, "flush"):
-                    stdout.flush()  # type: ignore
+                    stdout.flush()
 
             if status != "success":
                 _logger.warning(u"Final process status is %s", status)
@@ -848,12 +861,12 @@ def main(argsl=None,                   # type: List[str]
             return 0
 
         except (validate.ValidationException) as exc:
-            _logger.error(u"Input object failed validation:\n%s", exc,
+            _logger.error(u"Input object failed validation:\n%s", Text(exc),
                           exc_info=args.debug)
             return 1
         except UnsupportedRequirement as exc:
             _logger.error(
-                u"Workflow or tool uses unsupported feature:\n%s", exc,
+                u"Workflow or tool uses unsupported feature:\n%s", Text(exc),
                 exc_info=args.debug)
             return 33
         except WorkflowException as exc:
@@ -863,15 +876,19 @@ def main(argsl=None,                   # type: List[str]
             return 1
         except Exception as exc:  # pylint: disable=broad-except
             _logger.error(
-                u"Unhandled error%s:\n  %s", try_again_msg, exc, exc_info=args.debug)
+                u"Unhandled error%s:\n  %s", try_again_msg, Text(exc), exc_info=args.debug)
             return 1
 
     finally:
         if args and runtimeContext and runtimeContext.research_obj \
-                and workflowobj:
+                and workflowobj and loadingContext:
             research_obj = runtimeContext.research_obj
-            prov_dependencies = prov_deps(workflowobj, document_loader, uri)
-            research_obj.generate_snapshot(prov_dependencies)
+            if loadingContext.loader is not None:
+                research_obj.generate_snapshot(prov_deps(
+                    workflowobj, loadingContext.loader, uri))
+            else:
+                _logger.warning("Unable to generate provenance snapshot "
+                    " due to missing loadingContext.loader.")
             if prov_log_handler is not None:
                 # Stop logging so we won't half-log adding ourself to RO
                 _logger.debug(u"[provenance] Closing provenance log file %s",
@@ -888,11 +905,11 @@ def main(argsl=None,                   # type: List[str]
         _logger.addHandler(defaultStreamHandler)
 
 
-def find_default_container(builder,                  # type: HasReqsHints
-                           default_container=None,   # type: Text
-                           use_biocontainers=None,  # type: bool
+def find_default_container(builder,                 # type: HasReqsHints
+                           default_container=None,  # type: Optional[Text]
+                           use_biocontainers=None,  # type: Optional[bool]
                           ):  # type: (...) -> Optional[Text]
-    """Default finder for default containers."""
+    """Find a container."""
     if not default_container and use_biocontainers:
         default_container = get_container_from_software_requirements(
             use_biocontainers, builder)
@@ -900,7 +917,7 @@ def find_default_container(builder,                  # type: HasReqsHints
 
 
 def run(*args, **kwargs):
-    # type: (...) -> None
+    # type: (*Any, **Any) -> None
     """Run cwltool."""
     signal.signal(signal.SIGTERM, _signal_handler)
     try:
